Unit WavPlay;
{***************************************************************************
** Unit to play WAV-format files from Turbo Pascal for DOS.               **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.home.ml.org                                               **
**                                                                         **
***************************************************************************}

interface

{
  This is used for transmitting wave-information to the calling programme
}
type WType = record
  SampleRate : Word;
  Resolution : Byte;
  Stereo : Boolean;
  Length : LongInt;
end;

{
  The only 2 procedures available to the calling programme.

  ** Playing a Wave-file
    All the calling programme has to do is :
      PlayWave ('TEST.WAV');
    and the appropriate file will be played through any SB-compatible card.
    Is this easy or what? I don't think it comes any easier!
    On second thought :  read the next bit.

  ** Stopping the currently playing Wave-file :
      StopWave;
    Anyone who has trouble doing that is on the wrong news-group and
    probably has an IQ below absolute 0.    <g>

}
procedure PlayWave (f : String);
procedure StopWave;

{Variables available to the calling programme : }
var Playing : Boolean; {True if currently playing a wave-file}
    WaveError : Byte; {If an error has occurred, contains the error code}
    WaveType : WType; {Contains wave-file information}

const FileNotFound = $01;
      NoCard = $02;
      NoStereoCard = $03;
      No16BitCard = $04;
      {
        Different error codes. Note that these are the only things
        the programme checks for. Additional errors might produce
        unexpected results
      }

implementation

{Needed for I/O}
uses Dos, Crt;

{Definition of chunk-formats}
type HeaderChunk = record
  RIFF : LongInt;
  NextChunkSize : LongInt;
end;
type WaveChunk = record
  WAVE : LongInt;
end;
type fmtHeader = record
  fmt : LongInt;
  fmtDataLength : LongInt;
end;
type fmtChunk = record
  WaveType : Word;
  Channels : Word;
  SampleRate : LongInt;
  BytesPerSecond : LongInt;
  BlockAlignment : Word;
  BitResolution : LongInt;
end;
type DataChunk = record
  data : LongInt;
  SoundLength : LongInt;
end;
type ThirtyTwoK = array [0..32767] of Byte;

{Local variables}
var Old13, Old15, OldEP : pointer; {Pointers to old routines}
    Base : Word; {The SB base address}
    SoundSegment : ^ThirtyTwoK; {Indicates the segment containing the sound buffer}
    SoundRead, {Indicates the amount of bytes already read}
    OverHead : LongInt; {If the amount of bytes to be played is not a
                         multiple of 16384 (most likely, it isn't),
                         this is used to store the difference}
    NextTimeFreeSoundSegment, Upper : Boolean;
    SoundFile : File; {What do you think?}
    Card, {Indicates the detected card}
    Channel : Byte; {
                      Stores the current DMA-channel :
                      either 1 (8-Bit) or 5 (16-Bit)
                    }
    DMAChnl8Bit,
    DMAChnl16Bit : Byte; {
                           These two are currently hard coded,
                           detection *might* be added later.
                         }
    Header : HeaderChunk;
    Wave : WaveChunk;
    fmtH : fmtHeader;
    fmt : fmtChunk;
    data : DataChunk; {
                        These are used to read in the different
                        wave chunks
                      }

const SB8 = 8; SB16 = 16;
      {Constants used for programming of the DMA-controller}
      DmaChannel : Array [0..7, 1..3] of Byte  =
        ( ($87, $0, $1), ($83, $2, $3), ($81, $4, $5), ($82, $6, $7),
         ($8F, $C0, $C2), ($8B, $C4, $C6), ($89, $C8, $CA), ($8A, $CC, $CE));
      DmaPort : Array [0..7, 1..3] of Byte  =
        ( ($A, $B, $C), ($A, $B, $C), ($A, $B, $C), ($A, $B, $C),
         ($D4, $D6, $D8), ($D4, $D6, $D8), ($D4, $D6, $D8), ($D4, $D6, $D8));
      {Chunk-identification values}
      rId = $46464952;  {RIFF}
      wId = $45564157;  {WAVE}
      fId = $20746D66;  {fmt }
      dId = $61746164;  {data}

procedure WriteDSP (value : byte);
 (*
  Writes a value to the DSP-chip on the SB
*)
begin
  while Port [base + $C] And $80 <> 0 do;
  Port [base + $C] := value;
end;

function ReadDSP : byte;
 (*
  Reads a value to the DSP-chip on the SB
*)
begin
  while Port [base + $E] and $80 = 0 do;
  ReadDSP := Port [base + $A];
end;

function SpeakerOn :  byte;
 (*
  Establishes the DSP<->Speaker connection, necessary for older cards.
*)
begin
  WriteDSP ($D1);
end;

function SpeakerOff :  byte;
 (*
  Discontinues the DSP<->Speaker connection, necessary for older cards.
*)
begin
  WriteDSP ($D3);
end;

procedure DMAStop;
 (*
  Stops playing the wave-file.
*)
begin
  Playing := false; {Set general variable to indicate no sound}
  WriteDSP ($D0);  {Function :  D0 Stop 8 bit DMA transfer}
  WriteDSP ($D5);  {Function :  D0 Stop 16 bit DMA transfer}
  {$I-}
    Close (SoundFile); {Close the soundfile}
  {$I+}
  if IOResult <> 0 then; {Clear Error flag}
  {Free the sound buffer}
  {FreeMem (SoundSegment);}
end;

function ResetDSP (b : word) : boolean;
 (*
  Resets the DSP so that it can be used for our own programme.
*)
begin
  Port [b + $6] := 1;
  Delay (10);
  Port [b + $6] := 0;
  Delay (10);
  if (Port [b + $E] And $80 = $80) And
     (Port [b + $A] = $AA) then
    ResetDSP := true
  else
    ResetDSP := false;
end;

procedure Playback (SoundSeg, SoundOfs, size : longint);
 (*
  This procedure sets up the DMA controller for DMA transfer.
  Then it programs the DSP chip to receive the transfer.
  Finally it initiates the transfer.
*)
var
  page, offset : longint;

begin
  {Calculate Offset and Page address of Wave-data}
  if fmt.BitResolution = 8 then begin
    offset := SoundSeg Shl 4 + SoundOfs;
    page := (SoundSeg + SoundOfs shr 4) shr 12;
  end else begin
    size := size shr 1;
    page := (SoundSeg + SoundOfs shr 4) shr 12;
    offset := (SoundSeg Shl 3 + SoundOfs shr 1) mod 65536;
  end;
  {Decrease size by one. This is necessary because the
  DMA controller sends one byte/word more than it is told to}
  {Setup DMA Controller for transfer}
  Port [DMAPort    [Channel, 1]] := 4 or (Channel and 3);
  if fmt.BitResolution = 16 then Port [$D8] := 0;
  Port [DMAPort    [Channel, 3]] := 0;
  Port [DMAPort    [Channel, 2]] := $48 or (Channel and 3);
  Port [DMAChannel [Channel, 2]] := Lo (offset);
  Port [DMAChannel [Channel, 2]] := Hi (offset);
  Port [DMAChannel [Channel, 1]] := page;
  Port [DMAChannel [Channel, 3]] := Lo (size);
  Port [DMAChannel [Channel, 3]] := Hi (size);
  Port [DMAPort    [Channel, 1]] := (Channel and 3);
  {Set DSP}
  if Card = SB8 then begin
    {Set up 8-bit card, sorry no stereo SBPRO support}
    WriteDSP ($14);
  end else begin
    {Set up 16-bit card}
    if fmt.BitResolution = 8 then begin
      {8-Bit file}
      WriteDSP ($C0);
      if fmt.Channels = 1 then WriteDSP ($00); {Mono}
      if fmt.Channels = 2 then WriteDSP ($20); {Stereo}
    end else begin
      {16-Bit file
      Perhaps this also needs to be changed}
      WriteDSP ($B0);
      if fmt.Channels = 1 then WriteDSP ($10); {Mono}
      if fmt.Channels = 2 then WriteDSP ($30); {Stereo}
    end;
  end;
  {Send the size of the transfer to the SB}
  WriteDSP (Lo (size));
  WriteDSP (Hi (size));
  {Set global variable to indicate playing sound}
  Playing := true;
end;

procedure ServiceIRQ (Flags, CS, IP, AX, BX, CX, DX, SI, DI, DS, ES, BP :  Word); interrupt;
 (*
  This procedure is called at the end of a DMA transfer. It starts the
  playing of the next portion of the wave-file and reads in another block.
*)
var
  b, t : Byte;

begin
  {relieve card}
  if Card = SB16 then begin
    Port [base + $4] := $82;
    t := Port [base + $5];
    if t and 1 = 1 then b := Port [base + $E]; { 8bit interrupt}
    if t and 2 = 2 then b := Port [base + $F]; {16bit interrupt}
  end else begin
    {8bit interrupt}
    b := Port [base + $E];
  end;
  {acknowledge hardware interrupt}
  Port [$20] := $20;
  {stop playing}
  Playing := false;
  {if NextTimeFreeSoundSegment = true then FreeMem (SoundSegment);}
  {
    The following is done when the remaining part of the file
    is less than 16K. It plays that part.
  }
  if OverHead>0 then begin
    if Upper then begin
      PlayBack (Seg (SoundSegment^), Ofs (SoundSegment^), OverHead);
    end else begin
      PlayBack (Seg (SoundSegment^), Ofs (SoundSegment^) + 16384, OverHead);
    end;
    Close (SoundFile);
    OverHead := 0;
    {
      The next time this routine is called, the sound buffer must
      be freed so that the memory it occupies is available to the
      calling programme.
    }
    NextTimeFreeSoundSegment := true;
  end;
  {
    If there is more than 16K to be played and/or read, it will
    be done in chunks of 16K.
  }
  if data.SoundLength-SoundRead>0 then begin
    if data.SoundLength-SoundRead>16384 then begin
      {Load into appropriate part of the buffer}
      if Upper then begin
        PlayBack (Seg (SoundSegment^), Ofs (SoundSegment^), 16384);
        BlockRead (SoundFile, SoundSegment^ [16384], 16384);
      end else begin
        PlayBack (Seg (SoundSegment^), Ofs (SoundSegment^) + 16384, 16384);
        BlockRead (SoundFile, SoundSegment^, 16384);
      end;
      {Update position indicators}
      inc (SoundRead, 16384);
      Upper := Not Upper;
    end else begin
      {
        Load in the last part of the Wave-file
        Then play it.
      }
      OverHead := data.SoundLength-SoundRead;
      if Upper then begin
        PlayBack (Seg (SoundSegment^), Ofs (SoundSegment^), 16384);
        BlockRead (SoundFile, SoundSegment^ [16384], Overhead);
      end else begin
        PlayBack (Seg (SoundSegment^), Ofs (SoundSegment^) + 16384, 16384);
        BlockRead (SoundFile, SoundSegment^, Overhead);
      end;
      inc (SoundRead, Overhead);
      Upper := Not Upper;
    end;
  end;
end;

procedure PlayWave (f : String);
var Play : Boolean;
begin
  {Return error if no sound card found}
  WaveError := 0;
  if base = 0 then begin
    WaveError := NoCard;
  end;
  {Initialize settings}
  NextTimeFreeSoundSegment := false;
  OverHead := 0;
  FileMode := 0; {Necessary to protect files and to access read-only files}
  DMAStop; {Stop any DMA-transfer that might be in progress}
  Play := true; {Assume it can be played}
  {Check for existence of file}
  Assign (SoundFile, F);
  {$I-}
    Reset (SoundFile, 1);
  {$I+}
  if IOResult = 0 then begin
    {it exists, so read in the header}
    BlockRead (SoundFile, Header, 8);
    {check for 'RIFF', if not found :  don't play}
    if Header.RIFF <> rId then Play := false;
    {read in wave header}
    BlockRead (SoundFile, Wave, 4);
    {check for 'WAVE', if not found :  don't play}
    if Wave.WAVE <> wId then Play := false;
    if Play = true then begin
      fmtH.fmt := 0;
      {read in fmt Header, checking for 'fmt '}
      while fmtH.fmt <> fId do begin
        {
          This actually goes searching through the entire file.
          It is inefficient, but was necessary for some wave files
          which had some other (to me :  unknown) chunks in between.
        }
        BlockRead (SoundFile, fmtH, 8);
        if fmtH.fmt <> fId then Seek (SoundFile, FilePos (SoundFile)-7);
      end;
    end;
    {read in wave chunk}
    BlockRead (SoundFile, fmt, fmtH.fmtDataLength);
    {Is it 8 or 16 bit?}
    fmt.BitResolution := fmt.BitResolution and $FF;
    if Play = true then begin
      data.data := 0;
      {read in data chunk, checking for 'data'}
      while data.data <> dId do begin
        {
          Same search procedure as above
        }
        BlockRead (SoundFile, data, 8);
        if data.data <> dId then Seek (SoundFile, FilePos (SoundFile)-7);
      end;
    end;
    {
      Some wave-files have an incorrect SoundLength. This makes sure
      that the SoundLength is never larger than actually fits in the
      wave file.
    }
    if data.SoundLength>FileSize (SoundFile)-FilePos (SoundFile)-1 then
      data.SoundLength := FileSize (SoundFile)-FilePos (SoundFile)-1;
    WaveType.Length := data.SoundLength;
    {
      The WaveLength (not SoundLength) indicates the number of Samples,
      not the number of bytes, so this needs to be adjusted for the
      number of channels (Mono/Stereo) and the bit-resolution
      (8/16-Bit)
    }
    if WaveType.Stereo = true then WaveType.Length := WaveType.Length shr 1;
    if WaveType.Resolution = 16 then WaveType.Length := WaveType.Length shr 1;
    if fmt.WaveType <> 1 then Play := false;
  end else begin
    {it doesn't exist, so return error}
    WaveError := FileNotFound;
    Play := false;
  end;
  {stereo file over 8bit card, generates and error}
  if (fmt.Channels = 2) and (Card <> SB16) then begin
    WaveError := NoStereoCard;
    Play := false;
  end;
  {16bit file over 8bit card, generates an error}
  if (fmt.BitResolution = 16) and (Card <> SB16) then begin
    WaveError := No16BitCard;
    Play := false;
  end;
  {file is valid, so we can start playing}
  if Play = true then begin
    {set DMAChannel}
    if fmt.BitResolution = 8 then Channel := DMAChnl8Bit;
    if fmt.BitResolution = 16 then Channel := DMAChnl16Bit;
    {
      Update global variables so that calling programs can identify
      the wave being played. Pretty useless for games though
    }
    WaveType.SampleRate := fmt.SampleRate;
    WaveType.Resolution := fmt.BitResolution;
    if fmt.Channels = 2 then WaveType.Stereo := true else WaveType.Stereo := false;
    SoundRead := 0;
    {Allocate 32K of memory to the sound buffer}
    New (SoundSegment);
    {If there was an error allocating memory, don't play.}
    if SoundSegment = nil then exit;
    {set sample rate}
    if Card = SB8 then begin
      {Original SB requires a special 'time-frame' computation}
      WriteDSP ($40);
      WriteDSP (256 - 1000000 div fmt.SampleRate);
    end else begin
      {SB16 just needs the samplerate. Much easier}
      WriteDSP ($41);
      WriteDSP (hi (fmt.SampleRate));
      WriteDSP (lo (fmt.SampleRate));
    end;
    {check length of file}
    if data.SoundLength>32768 then begin
      {must be played in parts}
      BlockRead (SoundFile, SoundSegment^, 32768);
      SoundRead := 32768;
      PlayBack (Seg (SoundSegment^), Ofs (SoundSegment^), 16384);
      Upper := false;
    end else begin
      {can be played at once}
      BlockRead (SoundFile, SoundSegment^, data.SoundLength);
      PlayBack (Seg (SoundSegment^), Ofs (SoundSegment^), data.SoundLength);
      SoundRead := data.SoundLength;
    end;
  end;
end;

procedure StopWave;
begin
  DMAStop;
end;

{$F+}
procedure ExitWavePlayer;
begin
  {Restores the ExitProc pointer to the original value}
  ExitProc := OldEP;
  {Stops any DMA-transfer that might be in progress}
  DMAStop;
  {Free interrupt vectors used to service IRQs}
  SetIntVec (13, Old13);
  SetIntVec (15, Old15);
  {Mask IRQs}
  Port [$21] := Port [$21] or (1 shl 5);
  Port [$21] := Port [$21] or (1 shl 7);
end;
{$F-}

begin
  Playing := false;
  Base := 0;
  {
    Detect base address.
    This only checks at address 220h and 240h.
  }
  if ResetDSP ($240) = true then Base := $240;
  if ResetDSP ($220) = true then Base := $220;
  {
    Detect card type.
    This doesn't distinguish between the original SB and the SBPRO.
  }
  WriteDSP ($E1);
  if ReadDSP >= 4 then Card := SB16 else Card := SB8;
  {
    Set DMAChannel.
    I don't know how to detect this, so I hardcoded the default values.
  }
  DMAChnl8Bit := 1;
  DMAChnl16Bit := 5;
  if Base <> 0 then begin
    {
      I am too lazy to detect for the IRQ being used, so I just
      'hooked' both IRQ 5 and 7, for they are most often used.
      If you want to detect for it, just set up several
      IRQ-servicing routines and do a small, 1-byte DMA-transfer and
      check which routine is called.
    }
    WriteDSP ($D1);
    {Save old IRQ5->INT13 vector}
    GetIntVec (13, Old13);
    {Save old IRQ7->INT15 vector}
    GetIntVec (15, Old15);
    {Set new IRQ5->INT13 vector}
    SetIntVec (13, Addr (ServiceIRQ));
    {Set new IRQ7->INT15 vector}
    SetIntVec (15, Addr (ServiceIRQ));
    {Enable IRQ5}
    Port [$21] := Port [$21] and not (1 shl 5);
    {Enable IRQ7}
    Port [$21] := Port [$21] and not (1 shl 7);
    {
      Save ExitProc pointer and set it to our own exit procedure.
      The ExitProc procedure is called after the main (calling)
      programme terminates. The main programme doesn't have to take
      care of resetting the IRQs and so on.
    }
    OldEP := ExitProc;
    ExitProc := Addr (ExitWavePlayer);
  end;
end.